// Archer logic

#include "ArcherCommon.as"
#include "ThrowCommon.as"
#include "Knocked.as"
#include "Hitters.as"
#include "RunnerCommon.as"
#include "ShieldCommon.as";	
#include "Help.as";
#include "BombCommon.as";

const int FLETCH_COOLDOWN = 45;
const int PICKUP_COOLDOWN = 15;
const int fletch_num_arrows = 1;
const int STAB_DELAY = 10;
const int STAB_TIME = 18;

void onInit( CBlob@ this )
{
	ArcherInfo archer;	  
	
	archer.stab_delay = 0;
	archer.fletch_cooldown = 0;
	archer.arrow_type = ArrowType::normal;
	this.set("archerInfo", @archer);

    this.set_s8( "charge_time", 0 );
    this.set_u8( "charge_state", ArcherParams::not_aiming );
    this.set_bool( "has_arrow", false );
    this.set_f32("gib health", -1.5f);
    this.Tag("player");
    this.Tag("flesh");		   

	this.set_Vec2f("inventory offset", Vec2f(0.0f, 122.0f));
    //no spinning
    this.getShape().SetRotationsAllowed(false);
    this.getSprite().SetEmitSound( "Entities/Characters/Archer/BowPull.ogg" );
    this.addCommandID("shoot arrow");
    this.addCommandID("pickup arrow");
																
	SetHelp( this, "help self hide", "archer", "Hide    $KEY_S$", "", 2 ); 
	SetHelp( this, "help self action2", "archer", "$Daggar$$Tree$Make arrow    $RMB$", "", 3 ); 

    setKnockable( this );
    for (uint i = 0; i < arrowTypeNames.length; i++) {
        this.addCommandID( "pick " + arrowTypeNames[i]);
    }

	this.getCurrentScript().runFlags |= Script::tick_not_attached;
	this.getCurrentScript().removeIfTag = "dead";
}

void onTick( CBlob@ this )
{
    Knocked(this);

	ArcherInfo@ archer;
	if (!this.get( "archerInfo", @archer )) {
		return;
	}

	CSprite@ sprite = this.getSprite();
    u8 charge_state = archer.charge_state;
    
	const bool stab = this.isKeyJustPressed( key_action2 );
	if (stab) 
	{
		// cancel charging
		if (charge_state != ArcherParams::stabbing && charge_state != ArcherParams::not_aiming)
		{	
			charge_state = ArcherParams::not_aiming;
			archer.charge_time = 0;
			sprite.SetEmitSoundPaused( true );
			sprite.PlaySound("PopIn.ogg");
		}
		else if (charge_state != ArcherParams::stabbing && !this.isKeyPressed( key_action1 ))
		{
			charge_state = ArcherParams::stabbing;
			archer.stab_delay = 0;
		}

		archer.charge_state = charge_state;		
	}

	if (charge_state == ArcherParams::stabbing)
	{
		u8 delay = archer.stab_delay;
		if (delay == STAB_DELAY) 
		{		
			// STAB
			sprite.PlaySound("SwordSlash.ogg");
			if (getNet().isServer()){
				Stab(this);
			}
		}

		if (delay > STAB_TIME) 
		{
			charge_state = ArcherParams::not_aiming;
			archer.charge_state = charge_state;		
		}
		
		if (this.isKeyJustPressed( key_action1 )) //cancel
		{
			delay = STAB_TIME + 1;
			charge_state = ArcherParams::not_aiming;
		}

		delay++;
		archer.stab_delay = delay;		
	}

	// vvvvvvvvvvvvvv CLIENT-SIDE ONLY vvvvvvvvvvvvvvvvvvv

	if (!getNet().isClient()) return;
	if (this.isInInventory())
		return;

	RunnerMoveVars@ moveVars;
	if (!this.get( "moveVars", @moveVars )) {
		return;
	}		  

	bool ismyplayer = this.isMyPlayer();
	bool hasarrow = archer.has_arrow;
	s8 charge_time = archer.charge_time;

	if (charge_state != ArcherParams::stabbing && !this.isKeyPressed( key_action2 ))
	{
		if ((getGameTime()+this.getNetworkID())%10 == 0)
		{
			hasarrow = hasArrows( this );

			if (!hasarrow) 
			{            
				// set back to default
				for (uint i = 0; i < ArrowType::count; i++) 
				{
					if (hasArrows( this, i)) 
					{
						archer.arrow_type = i;
						break;
					}
				}
			}
		}
		
		archer.stab_delay = 0;		

		if (this.isKeyPressed(key_action1))
		{
			moveVars.walkFactor *= 0.6f;
			moveVars.canVault = false;

			if (ismyplayer && hasarrow && getCamera() !is null)
			{
				getCamera().mousecamstyle = 1;
			}

			const bool just_action1 = this.isKeyJustPressed(key_action1);

		//	printf("charge_state " + charge_state );

			if ((just_action1 || this.wasKeyPressed(key_action2) && !this.isKeyPressed(key_action2)) &&
				(charge_state == ArcherParams::not_aiming || charge_state == ArcherParams::fired))
			{
				charge_state = ArcherParams::readying;
				hasarrow = hasArrows( this );

				if (!hasarrow)
				{
					charge_state = ArcherParams::no_arrows;
					charge_time = 0;

					if (this.isMyPlayer() && !this.wasKeyPressed(key_action1)) { // playing annoying no ammo sound
						Sound::Play("Entities/Characters/Sounds/NoAmmo.ogg");
					}

					// set back to default
					archer.arrow_type = ArrowType::normal;
				}
				else
				{
					if (ismyplayer)
					{		 		
						if (just_action1)
						{
							const u8 type = archer.arrow_type;
							if (type == ArrowType::fire) {
								sprite.PlaySound( "SparkleShort.ogg" );
							} else if (type == ArrowType::water) {
								sprite.PlaySound( "/WaterBubble" );
							}
						}
					}
					
					sprite.RewindEmitSound();
					sprite.SetEmitSoundPaused( false );

					if (!ismyplayer) { // lower the volume of other players charging  - ooo good idea
						sprite.SetEmitSoundVolume( 0.5f );
					}
				}
			}
			else if (charge_state == ArcherParams::readying)
			{
				charge_time++;

				if(charge_time > ArcherParams::ready_time)
				{
					charge_time = 1;
					charge_state = ArcherParams::charging;
				}
			}
			else if (charge_state == ArcherParams::charging)
			{
				charge_time++;

				if(charge_time >= ArcherParams::shoot_period)
				{
					charge_state = ArcherParams::decharging;
					sprite.SetEmitSoundPaused( true );
				}
			}
			else if (charge_state == ArcherParams::decharging)
			{
				charge_time--;

				if(charge_time == 0)
				{
					charge_state = ArcherParams::charging;
					sprite.RewindEmitSound();
					sprite.SetEmitSoundPaused( false );
				}
			}
			else if (charge_state == ArcherParams::no_arrows)
			{
				if(charge_time < ArcherParams::ready_time) {
					charge_time++;
				}
			}
		}
		else
		{
			if (charge_state > ArcherParams::readying)
			{
				if (charge_state < ArcherParams::fired)
				{
					//time to fire!
					if (hasarrow && canSendArrow(this) ) // client-logic
					{
						f32 arrowspeed;

						if (charge_time < (ArcherParams::shoot_period/3)) {
							arrowspeed = ArcherParams::shoot_max_vel * (1.0f/3.0f);
						}
						else if (charge_time < (2*ArcherParams::shoot_period/3)) {
							arrowspeed = ArcherParams::shoot_max_vel * (2.0f/3.0f);
						}
						else {
							arrowspeed = ArcherParams::shoot_max_vel;
						}

						ShootArrow( this, this.getPosition() + Vec2f(0.0f,-2.0f), this.getAimPos()+Vec2f(0.0f,-2.0f), arrowspeed, archer.arrow_type );
					}

					charge_time = ArcherParams::fired_time;
					charge_state = ArcherParams::fired;
				}
				else //fired..
				{
					charge_time--;

					if (charge_time <= 0) {
						charge_state = ArcherParams::not_aiming;
						charge_time = 0;
					}
				}
			}
			else {
				charge_state = ArcherParams::not_aiming;    //set to not aiming either way
				charge_time = 0;
			}

			sprite.SetEmitSoundPaused( true );
		}
	}

	// safe disable bomb light

	if (this.wasKeyPressed(key_action1) && !this.isKeyPressed(key_action1) )
	{
		const u8 type = archer.arrow_type;
		if (type == ArrowType::bomb) {
			BombFuseOff( this );
		}
	}

	// my player!

    if ( ismyplayer )
    {			
		// activate/throw

        if (this.isKeyJustPressed(key_action3))
        {
			client_SendThrowOrActivateCommand( this );	  
        }
    		
		// pick up arrow

		if (archer.fletch_cooldown > 0) {
			archer.fletch_cooldown--;
		}

		// pickup from ground

        if (archer.fletch_cooldown == 0 && this.isKeyPressed(key_action2))
        {
            if (getPickupArrow( this ) !is null /*|| canPickSpriteArrow( this, false ) doesnt work*/) // pickup arrow from ground
            {
                this.SendCommand( this.getCommandID("pickup arrow") );
				archer.fletch_cooldown = PICKUP_COOLDOWN;
            }
        }
    }   

	archer.charge_time = charge_time;
	archer.charge_state = charge_state;
	archer.has_arrow = hasarrow;
}

bool canSendArrow( CBlob@ this )
{
	return (this.isMyPlayer() || this.getPlayer() is null || this.getPlayer().isBot());
}

void ShootArrow( CBlob @this, Vec2f arrowPos, Vec2f aimpos, f32 arrowspeed, const u8 arrow_type )
{
    if (canSendArrow(this)) 
	{ // player or bot
		f32 randomInn = 0.0f -2.0f + (( f32(Random(2048)) / 2048.0f) * 4.0f);
		Vec2f arrowVel = (aimpos- arrowPos).RotateBy(randomInn,Vec2f(0,0));
		arrowVel.Normalize();
		arrowVel *= arrowspeed;
		//print("arrowspeed " + arrowspeed);
		CBitStream params;
		params.write_Vec2f( arrowPos );
		params.write_Vec2f( arrowVel );
		params.write_u8( arrow_type );
								
		this.SendCommand( this.getCommandID("shoot arrow"), params );
	}
}

CBlob@ getPickupArrow( CBlob@ this )
{
	CBlob@[] blobsInRadius;
	if (this.getMap().getBlobsInRadius( this.getPosition(), this.getRadius()*1.5f, @blobsInRadius ))
	{
		for (uint i = 0; i < blobsInRadius.length; i++)
		{
			CBlob @b = blobsInRadius[i];
			if (b.getName() == "arrow")
			{
				return b;
			}
		}
	}
    return null;
}

bool canPickSpriteArrow( CBlob@ this, bool takeout )
{
	CBlob@[] blobsInRadius;
	if (this.getMap().getBlobsInRadius( this.getPosition(), this.getRadius()*1.5f, @blobsInRadius ))
	{
		for (uint i = 0; i < blobsInRadius.length; i++)
		{
			CBlob @b = blobsInRadius[i];
			{
				CSprite@ sprite = b.getSprite();
				if (sprite.getSpriteLayer("arrow") !is null)
				{
					if (takeout)
						sprite.RemoveSpriteLayer("arrow");
					return true;
				}
			}
		}
	}
	return false;
}

void onCommand( CBlob@ this, u8 cmd, CBitStream @params )
{
    if (cmd == this.getCommandID("shoot arrow"))
    {
        Vec2f arrowPos = params.read_Vec2f();
        Vec2f arrowVel = params.read_Vec2f();
        u8 arrowType = params.read_u8();
		ArcherInfo@ archer;
		if (!this.get( "archerInfo", @archer )) {
			return;
		}
		archer.arrow_type = arrowType;

		// return to normal arrow - server didnt have this synced
		if (!hasArrows( this, arrowType )) {
			return;
		}

        if (getNet().isServer())
        {
            CBlob@ arrow = server_CreateBlobNoInit( "arrow" );	
            if (arrow !is null)
            {
				// fire arrow?
				arrow.set_u8("arrow type", arrowType );
				arrow.Init();

				arrow.IgnoreCollisionWhileOverlapped( this );
                arrow.SetDamageOwnerPlayer( this.getPlayer() );
				arrow.server_setTeamNum( this.getTeamNum() );
				arrow.setPosition( arrowPos );
                arrow.setVelocity( arrowVel );
            }
        }

        this.getSprite().PlaySound( "Entities/Characters/Archer/BowFire.ogg" );
        this.TakeBlob( arrowTypeNames[ arrowType ], 1 );

		archer.fletch_cooldown = FLETCH_COOLDOWN; // just don't allow shoot + make arrow
    }
    else if (cmd == this.getCommandID("pickup arrow"))
    {
        CBlob@ arrow = getPickupArrow( this );
		bool spriteArrow = canPickSpriteArrow( this, false );
        if (arrow !is null || spriteArrow)
        {
			if(arrow !is null)
			{
				ArcherInfo@ archer;
				if (!this.get( "archerInfo", @archer )) {
					return;
				}
				const u8 arrowType = archer.arrow_type;
				if(arrowType == ArrowType::bomb)
				{
					arrow.set_u16("follow", 0); //this is already synced, its in command.
					arrow.setPosition(this.getPosition());
					return;
				}
			}
			
			CBlob@ mat_arrows = server_CreateBlob( "mat_arrows", this.getTeamNum(), this.getPosition() );
			if (mat_arrows !is null)
			{
				mat_arrows.server_SetQuantity(fletch_num_arrows);
				mat_arrows.Tag("do not set materials");
				this.server_PutInInventory( mat_arrows );

				if (arrow !is null) {
					arrow.server_Die();
				}
				else{
					canPickSpriteArrow( this, true );
				}
			}
			this.getSprite().PlaySound( "Entities/Items/Projectiles/Sounds/ArrowHitGround.ogg" );
        }
    }
    else if (cmd == this.getCommandID("cycle"))  //from standardcontrols
	{
		// cycle arrows
		ArcherInfo@ archer;
		if (!this.get( "archerInfo", @archer )) {
			return;
		}
		u8 type = archer.arrow_type;

		uint count = 0;
		while(count < arrowTypeNames.length)
		{
			type++;
			count++;
			if (type >= arrowTypeNames.length)
				type = 0;
			if (this.getBlobCount( arrowTypeNames[type] ) > 0)
			{
				archer.arrow_type = type;
				if (this.isMyPlayer())
				{
					Sound::Play("/CycleInventory.ogg");
				}
				break;
			}
		}
	}
	else
    {
		ArcherInfo@ archer;
		if (!this.get( "archerInfo", @archer )) {
			return;
		}
        for (uint i = 0; i < arrowTypeNames.length; i++)
        {
            if (cmd == this.getCommandID( "pick " + arrowTypeNames[i]))
            {
                archer.arrow_type = i;
                break;
            }
        }
    }
}

// arrow pick menu

void onCreateInventoryMenu( CBlob@ this, CBlob@ forBlob, CGridMenu @gridmenu )
{
    if (arrowTypeNames.length == 0) {
        return;
    }

    this.ClearGridMenusExceptInventory();
    Vec2f pos( gridmenu.getUpperLeftPosition().x + 0.5f*(gridmenu.getLowerRightPosition().x - gridmenu.getUpperLeftPosition().x),
               gridmenu.getUpperLeftPosition().y - 32 * 1 - 2*24 );
    CGridMenu@ menu = CreateGridMenu( pos, this, Vec2f( arrowTypeNames.length, 2 ), "Current arrow" );

	ArcherInfo@ archer;
	if (!this.get( "archerInfo", @archer )) {
		return;
	}
	const u8 arrowSel = archer.arrow_type;

    if (menu !is null)
    {
		menu.deleteAfterClick = false;

        for (uint i = 0; i < arrowTypeNames.length; i++)
        {
            string matname = arrowTypeNames[i];
            CGridButton @button = menu.AddButton( arrowIcons[i], arrowNames[i], this.getCommandID( "pick " + matname) );

            if (button !is null)
            {
				bool enabled = this.getBlobCount( arrowTypeNames[i] ) > 0;
                button.SetEnabled( enabled );
				button.selectOneOnClick = true;
			
				//if (enabled && i == ArrowType::fire && !hasReqs(this, i))
				//{
				//	button.hoverText = "Requires a fire source $lantern$";
				//	//button.SetEnabled( false );
				//}
			
                if (arrowSel == i) {
                    button.SetSelected(1);
                }
			}				
        }
    }
}

// auto-switch to appropriate arrow when picked up
void onAddToInventory( CBlob@ this, CBlob@ blob )
{
	string itemname = blob.getName();
	if (this.isMyPlayer())
	{
		for (uint j = 0; j < arrowTypeNames.length; j++)
		{	  
			printf(" " + itemname + " " + arrowTypeNames[j] );
			if (itemname == arrowTypeNames[j])
			{
				SetHelp( this, "help self action", "archer", "Fire arrow   $KEY_HOLD$$LMB$", "", 3 ); 
				if (this.getInventory().getItemsCount() > 1) {
					SetHelp( this, "help inventory", "archer", "$Help_Arrow1$$Swap$$Help_Arrow2$         $KEY_TAP$$KEY_F$", "", 2 );
				}
				break;
			}
		}
	}

	CInventory@ inv = this.getInventory();
	if (inv.getItemsCount() == 0)
	{			
		ArcherInfo@ archer;
		if (!this.get( "archerInfo", @archer )) {
			return;

		}		

		for (uint i = 0; i < arrowTypeNames.length; i++)
		{
			if (itemname == arrowTypeNames[i]) {
				archer.arrow_type = i;
			}
		}
	}
}


void Stab( CBlob@ this )
{	   				
	Vec2f blobPos = this.getPosition();
	Vec2f aimPos = this.getAimPos();
	Vec2f aimDir = aimPos - blobPos;
	f32 aimangle = aimDir.Angle();
	Vec2f pos = blobPos - Vec2f(2,0).RotateBy(-aimangle);
	f32 attack_distance = this.getRadius() + 8.0f;
	f32 damage;
	f32 radius = this.getRadius();
	CMap@ map = this.getMap();

	HitInfo@[] hitInfos;  
	if (map.getHitInfosFromArc( pos, -aimangle, 180.0f, attack_distance, this, @hitInfos ))
	{
		//HitInfo objects are sorted, first come closest hits
		for (uint i = 0; i < hitInfos.length; i++)
		{
			HitInfo@ hi = hitInfos[i];
			CBlob@ b = hi.blob;
			if (b !is null) // blob
			{						  
				const bool large = b.getMass() > 500;
				if (b.getTeamNum() == this.getTeamNum() && !b.hasTag("dead")) { // no TK
					if (large) return;
					else continue;
				}
				if (b.hasTag("invincible"))	{
					continue;
				}

				Vec2f otherPos = b.getPosition();
				if (b.getHealth() > 0.0f && b.hasTag("flesh") && !b.hasTag("dead") &&
					((otherPos.x > blobPos.x && !this.isFacingLeft() && !b.isFacingLeft()) ||
					 (otherPos.x < blobPos.x && this.isFacingLeft() && b.isFacingLeft())
					)
					)
				{
					damage = 3.0f;
				}
				else
				{
					damage = 0.25f;
				}

				Vec2f velocity = otherPos - pos;
				this.server_Hit( b, hi.hitpos, velocity, damage, Hitters::stab, true); // server_Hit() is server-side only

				if (large) return;
			}
		}
	}	   
}

void onHitBlob( CBlob@ this, Vec2f worldPoint, Vec2f velocity, f32 damage, CBlob@ hitBlob, u8 customData )
{
	if (customData == Hitters::stab)
	{
		if (damage > 0.0f)
		{
			
			// fletch arrow
			if ( hitBlob.hasTag("tree") )	// make arrow from tree
			{
				if (getNet().isServer())
				{
					CBlob@ mat_arrows = server_CreateBlob( "mat_arrows", this.getTeamNum(), this.getPosition() );
					if (mat_arrows !is null)
					{
						mat_arrows.server_SetQuantity(fletch_num_arrows);
						mat_arrows.Tag("do not set materials");
						this.server_PutInInventory( mat_arrows );
					}
				}
				this.getSprite().PlaySound( "Entities/Items/Projectiles/Sounds/ArrowHitGround.ogg" );
			}
			else
				this.getSprite().PlaySound("KnifeStab.ogg");
		}

		if (blockAttack(hitBlob, velocity, 0.0f))
		{
			this.getSprite().PlaySound("/Stun");
			SetKnocked( this, 30 );
		}	
	}
}
